from astropy import units
from astropy.coordinates import EarthLocation
from astropy.time import Time as EarthTime
from copy import deepcopy
from lunarsky import MoonLocation, Time as LunarTime
import matplotlib.pyplot as plt
import numpy as np
import warnings

from . import dpss
from .rotations import rot_coords
from .healpix import Alm, grid2healpix, healpix2lonlat, map2alm


class Simulator:
    def __init__(
        self,
        beam,
        sky,
        obs_loc,
        t_start,
        moon=True,
        t_end=None,
        N_times=None,
        delta_t=None,
        frequencies=None,
        horizon=None,
        lmax=16,
    ):
        """
        Simulator class. Prepares and runs simulations.
        """
        self.lmax = lmax
        self.moon = moon
        # set up frequencies to run the simulation at
        if frequencies is None:
            frequencies = sky.frequencies
        self.frequencies = frequencies
        if moon:
            loc_object = MoonLocation
            self.sim_coords = "mcmf"  # simulation coordinate system
            Time = LunarTime  # define default time class
        else:
            loc_object = EarthLocation
            self.sim_coords = "equatorial"
            Time = EarthTime

        if isinstance(obs_loc, loc_object):
            self.loc = obs_loc
        else:
            lat, lon, alt = obs_loc
            self.loc = loc_object(
                lat=lat * units.deg,
                lon=lon * units.deg,
                height=alt * units.m,
            )

        self.t_start = Time(t_start, location=self.loc, scale="utc")
        if delta_t is not None:
            try:
                delta_t = delta_t.to_value(units.s)
            except AttributeError:
                warnings.warn(
                    "No units specified for delta_t, assuming seconds.",
                    UserWarning,
                )
                delta_t = delta_t
        if t_end is None:
            dt = np.arange(N_times) * delta_t
        else:
            t_end = Time(t_end, location=self.loc)
            total_time = (t_end - self.t_start).to_value(units.s)
            if delta_t is not None:
                dt = np.arange(0, total_time + delta_t, delta_t)
                N_times = len(dt)
            else:
                dt = np.linspace(0, total_time, N_times)
        self.dt = dt
        self.N_times = N_times
        beam = deepcopy(beam)
        if beam.alm is None:
            # apply horizon mask and initialize beam
            beam.horizon_cut(horizon=horizon)
            self.beam = beam
            self.beam_alm()  # compute alms in ra/dec
        else:
            self.beam = beam
        # initialize sky
        self.sky = Alm.from_healpix(sky, lmax=self.lmax)
        if self.sky.coords.lower() != self.sim_coords:
            self.sky.switch_coords(self.sim_coords)

    def beam_alm(self, nside=128):
        """
        Get the alm's of the beam in the equatorial/MCMF coordinate system.
        """
        # get lon/lat in sim coordinates at healpix centers
        lon, lat = healpix2lonlat(nside)

        # get corresponding theta/phi in topocentric coords
        za = np.pi / 2 - np.deg2rad(lat)
        az = np.deg2rad(lon)
        theta, phi = rot_coords(
            za,
            az,
            self.sim_coords,
            self.beam.coords.lower(),
            time=self.t_start,
            loc=self.loc,
            lonlat=False,
        )

        pixel_centers = np.array([theta, phi]).T
        # get healpix map
        hp_maps = grid2healpix(
            self.beam.data,
            nside,
            self.beam.theta,
            self.beam.phi,
            pixel_centers=pixel_centers,
        )
        self.beam.alm = map2alm(hp_maps, self.lmax)
        self.beam.coords = self.sim_coords

    def compute_dpss(self, **kwargs):
        # generate the set of target frequencies (subset of all freqs)
        x = np.unique(
            np.concatenate(
                (
                    self.beam.frequencies,
                    self.frequencies,
                ),
                axis=None,
            )
        )

        self.design_matrix = dpss.dpss_op(x, **kwargs)
        self.beam.coeffs = dpss.freq2dpss(
            self.beam.alm,
            self.beam.frequencies,
            self.frequencies,
            self.design_matrix,
        )

    def run(self, dpss=True, **dpss_kwargs):
        """
        Compute the convolution for a range of times.
        """
        if self.moon:
            world = "moon"
        else:
            world = "earth"
        # the rotation phases
        phases = self.sky.rotate_alm_time(self.dt, world=world)
        if dpss:
            self.compute_dpss(**dpss_kwargs)
            # get the sky coefficients at each time
            rot_sky_coeffs = np.expand_dims(self.sky.alm, axis=0) * phases
            # m = 0 modes
            res = (
                rot_sky_coeffs[:, :, : self.lmax + 1].real
                @ self.beam.coeffs[:, : self.lmax + 1].T.real
            )
            # for m != 0 we must account for +m and -m
            res += 2 * np.real(
                rot_sky_coeffs[:, :, self.lmax + 1 :]
                @ self.beam.coeffs[:, self.lmax + 1 :].T.conj()
            )

            waterfall = np.einsum("ijk, jk -> ij", res, self.design_matrix)

        else:
            # add time dimensions and rotate the sky alms
            rot_sky_coeffs = np.expand_dims(self.sky.alm, axis=0) * phases
            beam_coeffs = np.expand_dims(self.beam.alm, axis=0)
            # m = 0 modes (already real)
            waterfall = np.einsum(
                "ijk, ijk -> ij",
                rot_sky_coeffs[:, :, : self.lmax + 1].real,
                beam_coeffs[:, :, : self.lmax + 1].real,
            )
            # for m != 0 we must account for +m and -m
            waterfall += 2 * np.real(
                np.einsum(
                    "ijk, ijk -> ij",
                    rot_sky_coeffs[:, :, self.lmax + 1 :],
                    beam_coeffs.conj()[:, :, self.lmax + 1 :],
                )
            )

        norm = self.beam.total_power.reshape(1, -1)
        self.waterfall = waterfall / norm

    def plot(
        self,
        figsize=None,
        extent=None,
        interpolation="none",
        aspect="auto",
        power=0,
    ):
        """
        Plot the result of the simulation.
        """
        plt.figure(figsize=figsize)
        if extent is None:
            extent = [
                self.frequencies.min(),
                self.frequencies.max(),
                self.dt[-1] / 3600,
                0,
            ]
        weight = self.frequencies**power
        plt.imshow(
            self.waterfall * weight.reshape(1, -1),
            extent=extent,
            aspect=aspect,
            interpolation=interpolation,
        )
        plt.colorbar(label="Temperature [K]")
        plt.xlabel("Frequency [MHz]")
        plt.ylabel(
            f"Hours since {self.t_start.to_value('iso', subfmt='date_hm')}"
        )
        plt.show()



from healpy import nside2npix
import numpy as np
import warnings

from .constants import Y00
from .healpix import grid2healpix


class Beam:
    def __init__(
        self,
        data,
        theta=None,
        phi=None,
        frequencies=None,
        alm=False,
        coords="topocentric",
    ):
        """
        Initialize beam object. The beam must be specified at a rectangular
        grid, that is, theta and phi must be coordinate axis of a grid.

        Parameters
        ----------
        data : array-like
            The power beam. Must have shape ([freqs,] theta, phi).
        theta : array-like
            Zenith angle(s) in radians. Must be in [0, pi].
        phi : array-like
            Azimuth angle(s) in radians. Must be in [0, 2*pi).
        frequencies : array-like (optional)
            The frequencies in MHz of the beam. Necessary if the beam is
            specified at more than one frequency.

        """
        data = np.array(data, copy=True)
        self.frequencies = np.ravel(frequencies).copy()
        self.nfreqs = self.frequencies.size
        if not alm:
            self.theta = np.ravel(theta).copy()
            self.phi = np.ravel(phi).copy()
            if self.theta.min() < 0 or self.theta.max() > np.pi:
                raise ValueError("Theta must be in the range [0, pi].")
            if self.phi.min() < 0 or self.phi.max() >= 2 * np.pi:
                raise ValueError("Phi must be in the range [0, 2pi).")
            if not (
                np.allclose(np.diff(self.theta), self.theta[1] - self.theta[0])
                and np.allclose(np.diff(self.phi), self.phi[1] - self.phi[0])
            ):
                raise ValueError(
                    "The data must be sampled on a rectangular grid."
                )
            data.shape = (self.nfreqs, theta.size, phi.size)
            self.data = data
            self.alm = None
        else:
            data.shape = (self.nfreqs, -1)
            self.data = None
            self.alm = data
        self.total_power = self.compute_total_power()  # before horizon cut
        self.coords = coords

    def compute_total_power(self, nside=128):
        """
        Compute the total integrated power in the beam at each frequency. This
        is a necessary normalization constant for computing the visibilities.
        It should be computed before applying the horizon cut in order to
        account for ground loss.
        """
        if self.data is not None:  # from grid
            healpix_beam = grid2healpix(
                self.data, nside, theta=self.theta, phi=self.phi
            )
            npix = nside2npix(nside)
            power = healpix_beam.sum(axis=-1) * 4 * np.pi / npix
        else:  # from alm
            power = self.alm[0, 0] * Y00 * 4 * np.pi
        return power

    def horizon_cut(self, horizon=None):
        """
        horizon : array-like (optional)
            An array indicating if a given (frequency/)theta/phi combination is
            above the horizon or not. Must have shape (theta, phi) or
            (freqs, theta, phi) if frequency dependent. The array will be
            multiplied by the antenna beam, hence 0 is intepreted as below the
            horizon and 1 is interpreted as above it. The elements must be in
            the range [0, 1].
        """
        # has no effect on alm beam
        if self.data is None:
            warnings.warn(
                "Horizon cut is not implemented for alm beams.",
                UserWarning,
            )
            return

        if horizon is None:
            horizon = np.ones_like(self.data)
            horizon[:, self.theta > np.pi / 2] = 0
        else:
            horizon = np.array(horizon, copy=True)
            horizon.shape = (-1, self.theta.size, self.phi.size)
            if horizon.min() < 0 or horizon.max() > 1:
                raise ValueError("Horizon elements must be in [0, 1].")
        self.data = self.data * horizon

    @classmethod
    def from_file(path):
        raise NotImplementedError

    def to_file(fname):
        raise NotImplementedError



import healpy as hp
import numpy as np

from croissant.sky import Sky


def test_power_law_map():
    nside = 32
    npix = hp.nside2npix(nside)
    freq = 25.0
    data = np.arange(npix)
    sky = Sky(data, frequencies=freq)
    freq_out = np.linspace(1, 50)
    beta = -2.5
    sky.power_law_map(freq_out, spectral_index=beta)
    assert np.allclose(sky.frequencies, freq_out)
    for pix in range(npix):  # check that there's a power law at each pixel
        temp = sky.data[:, pix]
        f_ratio = freq_out / freq
        f_ratio.shape = (1, -1)
        pl = data[pix] * f_ratio**beta
        assert np.allclose(temp, pl)



import numpy as np
import pytest

from croissant.beam import Beam


def test_beam_init():
    # initialize from grid
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    theta = np.linspace(0, np.pi, 181)
    phi, theta = np.meshgrid(phi, theta, sparse=True)
    # mock data
    data = np.sin(theta) ** 2 * np.cos(phi) ** 2
    beam = Beam(data, theta=theta, phi=phi)
    assert np.allclose(beam.data, data)
    assert beam.alm is None
    assert beam.data.shape == (1, theta.size, phi.size)

    # with phi outside range
    phi -= 2 * np.pi
    data = np.sin(theta) ** 2 * np.cos(phi) ** 2
    with pytest.raises(ValueError):
        Beam(data, theta=theta, phi=phi)

    # with irregular grid
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    theta = np.linspace(0, np.pi, 181)
    phi, theta = np.meshgrid(phi, theta, sparse=True)
    # remove a point from theta to make it irregular
    theta = np.delete(theta, theta.size // 2, axis=0)
    data = np.sin(theta) ** 2 * np.cos(phi) ** 2
    with pytest.raises(ValueError):
        Beam(data, theta=theta, phi=phi)

    # initialize from alm
    size = 10
    alm = np.arange(size) + 1j * np.arange(size) ** 2
    beam = Beam(alm, alm=True)
    assert beam.data is None
    assert np.allclose(beam.alm, alm)
    assert beam.alm.shape == (1, size)

    # initialize at a range of frequencies
    freqs = np.linspace(1, 50, 50)
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    theta = np.linspace(0, np.pi, 181)
    theta, freqs, phi = np.meshgrid(theta, freqs, phi, sparse=True)
    data = freqs**2 * np.sin(theta) ** 2 * np.cos(phi) ** 2
    beam = Beam(data, theta=theta, phi=phi, frequencies=freqs)
    assert np.allclose(beam.data, data)
    assert beam.alm is None
    assert beam.data.shape == (freqs.size, theta.size, phi.size)
    assert np.allclose(beam.frequencies, np.squeeze(freqs))


def test_compute_total_power():
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    theta = np.linspace(0, np.pi, 181)
    # make a beam that is 1 everywhere so total power is 4pi:
    data = np.ones((theta.size, phi.size))
    beam = Beam(data, theta, phi)
    assert np.allclose(beam.compute_total_power(), 4 * np.pi)
    # dipole beam
    frequencies = np.linspace(1, 50, 50).reshape(-1, 1, 1)
    data = np.cos(theta.reshape(1, -1, 1)) ** 2 * frequencies**2
    data = np.repeat(data, phi.size, axis=2)
    beam = Beam(data, theta=theta, phi=phi, frequencies=frequencies)
    power = beam.compute_total_power()
    expected_power = 4 * np.pi / 3 * frequencies**2
    assert np.allclose(power, expected_power.ravel())


def test_horizon_cut():
    # generate mock data on grid
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    theta = np.linspace(0, np.pi, 181)
    frequencies = np.linspace(1, 50, 50).reshape(-1, 1, 1)
    # make a beam that's constant in space, but varies in frequency
    data = np.ones((1, theta.size, phi.size)) * frequencies**2
    beam = Beam(data, theta=theta, phi=phi, frequencies=frequencies)

    # default horizon, i.e. cuts everywhere at theta > np.pi/2
    beam.horizon_cut(horizon=None)
    below_horizon = theta > np.pi / 2
    assert np.allclose(beam.data[:, below_horizon, :], 0)
    assert np.allclose(
        beam.data[:, ~below_horizon, :], data[:, ~below_horizon, :]
    )

    # custom horizon
    beam = Beam(data, theta, phi, frequencies=frequencies)
    horizon = np.ones(data.shape)
    th_cut = theta > np.pi / 2 - np.pi / 8
    ph_cut = phi < np.pi
    horizon[:, th_cut, :] = 0
    horizon[:, :, ph_cut] = 0
    beam.horizon_cut(horizon=horizon)
    assert np.allclose(beam.data[:, th_cut, :], 0)
    assert np.allclose(beam.data[:, :, ph_cut], 0)
    assert np.allclose(
        beam.data[:, ~th_cut, :][:, :, ~ph_cut],
        data[:, ~th_cut, :][:, :, ~ph_cut],
    )



import numpy as np

# nside's for which healpix has computed pixel weights:
PIX_WEIGHTS_NSIDE = [32, 64, 128, 256, 512, 1024, 2048, 4096]

# sidereal days in seconds
# https://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html
sidereal_day_earth = 23.9345 * 3600
# https://nssdc.gsfc.nasa.gov/planetary/factsheet/moonfact.html
sidereal_day_moon = 655.720 * 3600

Y00 = 1 / np.sqrt(4 * np.pi)  # the 0,0 spherical harmonic function



import healpy as hp
import numpy as np
from scipy.interpolate import RectSphereBivariateSpline

from . import rotations, constants


def healpix2lonlat(nside, pix=None):
    """
    Compute the longtitudes and latitudes of the pixel centers of a healpix
    map.

    Parameters
    ----------
    nside : int
        The nside of the healpix map.
    pix : array-like (optional)
        Which pixels to get the longtitudes and latitudes of. Defaults to all
        pixels.

    Returns
    -------
    lon : scalar or np.ndarray
        The longtitude(s) in degrees. Range: [0, 360).
    lat : scalar or np.ndarray
        The latitude(s) in degrees. Range: [-90, 90].

    """
    if pix is None:
        pix = np.arange(hp.nside2npix(nside))
    lon, lat = hp.pix2ang(nside, pix, nest=False, lonlat=True)
    return lon, lat


def grid_interp(data, theta, phi, to_theta, to_phi):
    """
    Interpolate on a sphere from specfied theta and phi. The data must be
    on a rectangular grid.

    Parameters
    ----------
    data : array-like
        The data to interpolate. The last two dimensions must be (theta, phi).
        Can optionally have a 0th dimmension (e.g. a frequency dimension).
    theta : 1d-array
        The polar angles (colatitudes) in radians. Must be regularly sampled
        and strictly increasing.
    phi : 1d-array
        The azimuthal angles in radians. Must be regularly sampled and strictly
        increasing. Must be in the interval [0, 2*pi).
    to_theta : array-like
        The polar angles to interpolate to in radians.
    to_phi : array-like
        The azimuthal angles to interpolate to in radians.

    Returns
    -------
    interp_data : np.ndarray
        The interpolated data.

    """
    theta = np.ravel(theta).copy()
    phi = np.ravel(phi).copy()
    data = np.array(data, copy=True).reshape(-1, theta.size, phi.size)

    # remove poles before interpolating
    pole_values = np.full((len(data), 2), None)
    northpole = theta[0] == 0
    southpole = theta[-1] == np.pi
    if northpole:
        theta = theta[1:]
        pole_values[:, 0] = data[:, 0, 0]
        data = data[:, 1:]
    if southpole:
        theta = theta[:-1]
        pole_values[:, 1] = data[:, -1, 0]
        data = data[:, :-1]

    interp_data = np.empty((len(data), to_theta.size))
    for i in range(len(data)):
        interp = RectSphereBivariateSpline(
            theta, phi, data[i], pole_values=pole_values[i]
        )
        interp_data[i] = interp(to_theta, to_phi, grid=False)
    return interp_data


def grid2healpix(data, nside, theta=None, phi=None, pixel_centers=None):
    """
    Transform data defined on a rectangular grid on a sphere to healpix map(s).
    To compute a healpix map in a different coordinate system, compute the
    pixel centers of the target coordinate system and set the keyword argument
    pixel_centers.

    Parameters
    ----------
    data : array-like
        The data to transform. The last two dimensions must be (theta, phi).
        It may have an optional 0th dimension to generate multiple maps at
        the same time.
    nside : int
        The nside of the output healpix map.
    theta : 1d-array (optional)
        The polar angles in radians. Must be in [0, pi]. Defaults to 1-degree
        sampling.
    phi : 1d-array (optional)
        The azimuthal angles in radians. Must be in [0, 2pi). Defaults to
        1-degree sampling.
    pixel_centers: 2d-array (optional)
        The centers of the pixels in radians. Must have shape (npix, 2) and be
        ordered like healpix RING pixels. The 0th column corresponds to theta
        and the 1st to phi.

    Returns
    -------
    hp_map : np.ndarray
        The healpix map(s) in RING order with shape (n_maps, n_pixels).

    """
    npix = hp.nside2npix(nside)
    if pixel_centers is not None:
        if np.shape(pixel_centers) != (npix, 2):
            raise ValueError(f"Shape must be ({npix}, 2).")
        pix_theta = pixel_centers[:, 0]
        pix_phi = pixel_centers[:, 1]
    else:
        lon, lat = healpix2lonlat(nside)
        pix_theta = np.pi / 2 - np.deg2rad(lat)
        pix_phi = np.deg2rad(lon)

    if theta is None:
        theta = np.linspace(0, np.pi, num=181)
    else:
        theta = np.array(theta, copy=True)

    if phi is None:
        phi = np.linspace(0, 2 * np.pi, num=360, endpoint=False)
    else:
        phi = np.array(phi, copy=True)

    hp_map = grid_interp(data, theta, phi, pix_theta, pix_phi)
    return hp_map


def map2alm(data, lmax):
    """
    Compute the spherical harmonics coefficents of a healpix map.
    """

    data = np.array(data)
    npix = data.shape[-1]
    nside = hp.npix2nside(npix)
    use_pix_weights = nside in constants.PIX_WEIGHTS_NSIDE
    use_ring_weights = not use_pix_weights
    kwargs = {
        "lmax": lmax,
        "mmax": lmax,
        "use_weights": use_ring_weights,
        "use_pixel_weights": use_pix_weights,
    }
    if data.ndim == 1:
        alm = hp.map2alm(data, **kwargs)
    else:
        alm = np.empty(
            (len(data), hp.Alm.getsize(lmax, mmax=lmax)), dtype=np.complex128
        )
        for i in range(len(data)):
            alm[i] = hp.map2alm(data[i], **kwargs)
    return alm


def alm2map(alm, nside=128, mmax=None):
    alm = np.array(alm, copy=True)
    if alm.ndim == 1:
        alm.shape = (1, -1)
    nfreqs = alm.shape[0]
    lmax = hp.Alm.getlmax(alm.shape[1], mmax=mmax)
    if mmax is None:
        mmax = lmax
    npix = hp.nside2npix(nside)
    hp_map = np.empty((nfreqs, npix))
    for i in range(nfreqs):
        map_i = hp.alm2map(alm[i], nside, lmax=lmax, mmax=mmax)
        hp_map[i] = map_i
    return np.squeeze(hp_map)


class HealpixMap:
    def __init__(
        self,
        data=None,
        nside=None,
        nested_input=False,
        frequencies=None,
        coords="galactic",
    ):
        """
        The base class for healpix maps. This is a wrapper that does a lot of
        healpy operations in parallel for a list of frequencies.
        It ensures that all maps have the right shapes and provdes an
        interpolation method.
        """
        if frequencies is not None:
            self.frequencies = np.atleast_1d(frequencies).copy()
        else:
            self.frequencies = None
        self.coords = coords

        if data is None:
            nested_input = False
            self.nside = nside
        else:
            data = np.array(data, copy=True, dtype=np.float64)
            data.shape = (np.size(self.frequencies), -1)
            npix = data.shape[-1]
            self.nside = hp.npix2nside(npix)

        if nside is not None and nside != self.nside:
            raise ValueError(
                (
                    "Mismatch between nside and data shape. The data has"
                    f" shape {data.shape} which suggests nside = {self.nside},"
                    f" but nside is set to {nside}."
                )
            )

        if self.nside is not None:
            hp.pixelfunc.check_nside(self.nside, nest=nested_input)

        if nested_input:
            ix = hp.nest2ring(self.nside, np.arange(self.npix))
            data = data[:, ix]

        self.data = data

    @property
    def npix(self):
        """
        Get the number of pixels of the map.
        """
        return hp.nside2npix(self.nside)

    @classmethod
    def from_alm(cls, alm_obj, nside=None):
        """
        Construct a healpy map class from an Alm object (defined below).
        """
        if nside is None:
            nside = (alm_obj.lmax + 1) // 3
        hp_map = alm_obj.hp_map(nside=nside)
        obj = cls(
            data=hp_map,
            nside=nside,
            nested_input=False,
            frequencies=alm_obj.frequencies,
            coords=alm_obj.coords,
        )
        return obj

    def ud_grade(self, nside_out, **kwargs):
        """
        Change the resolution of the healpy map to nside_out.

        Note: The nside in and out must be valid for nested maps since
        this conversion is being done under the hood.
        """
        hp.pixelfunc.check_nside(self.nside, nest=True)
        hp.pixelfunc.check_nside(nside_out, nest=True)

        new_map = hp.ud_grade(self.data, nside_out, **kwargs)
        self.data = new_map
        self.nside = nside_out

    def switch_coords(self, to_coords):
        rotated_map = rotations.rotate_map(
            self.data, from_coords=self.coords, to_coords=to_coords
        )
        self.data = rotated_map
        self.coords = to_coords

    def alm(self, lmax=None):
        """
        Compute the spherical harmonics coefficents of the map.
        """
        if lmax is None:
            lmax = 3 * self.nside - 1
        return map2alm(self.data, lmax)

    def plot(self, frequency=None, **kwargs):
        """
        Simple plotter of healpix maps. Can plot in several projections,
        including ``mollweide'', ``cartesian'' and ``polar''.
        """
        if self.data.ndim == 2 and self.frequencies is None:
            _m = self.data[0]
        else:
            _m = self.data
        m = kwargs.pop("m", _m)
        title = None
        if self.frequencies is not None:
            if frequency is None:
                raise ValueError("Must specify which frequency to plot.")
            else:
                f_idx = np.argmin(np.abs(self.frequencies - frequency))
                f_to_plot = self.frequencies[f_idx]
                title = kwargs.pop("title", f"Frequency = {f_to_plot:.0f} MHz")
                m = self.data[f_idx]
        _ = hp.projview(m=m, title=title, **kwargs)


class Alm(hp.Alm):
    def __init__(
        self, alm=None, lmax=None, frequencies=None, coords="galactic"
    ):
        """
        Base class for spherical harmonics coefficients.

        Alm can be indexed with [freq_index, ell, emm] to get the
        coeffiecient corresponding to the given frequency index, and values of
        ell and emm. The frequencies can be index in the usual numpy way and
        may be 0 if the alms are specified for only one frequency.

        """
        if alm is None and lmax is None:
            raise ValueError("Specify at least one of lmax and alm.")

        self.frequencies = np.ravel(frequencies).copy()
        if alm is None:
            self.lmax = lmax
            self.all_zero()
        elif lmax is None:
            alm = np.array(alm, copy=True, dtype=np.complex128)
            self.alm = alm.reshape(self.frequencies.size, -1)
            self.lmax = super().getlmax(alm.shape[1])
        else:
            self.lmax = lmax
            alm = np.array(alm, copy=True, dtype=np.complex128)
            self.alm = alm.reshape(*self.shape)

        self.coords = coords

    def __setitem__(self, key, value):
        """
        Set the value of the alm given the frequency index and the values of
        ell and emm.
        """
        # if alm only has one frequency, it doesn't matter if the freq_idx is
        # not specified:
        if self.shape[0] == 1 and len(key) == 2:
            ell, emm = key
            key = [0, ell, emm]
        if len(key) != 3:
            raise IndexError(
                f"Key has length {len(key)}, but must have length 3 to specify"
                " frequency index, ell, and emm."
            )
        freq_idx, ell, emm = key
        ix = self.getidx(ell, emm)
        self.alm[freq_idx, ix] = value

    def __getitem__(self, key):
        # if alm only has one frequency, it doesn't matter if the freq_idx is
        # not specified:
        if self.shape[0] == 1 and len(key) == 2:
            ell, emm = key
            key = [0, ell, emm]
        if len(key) != 3:
            raise IndexError(
                f"Key has length {len(key)}, but must have length 3 to specify"
                " frequency index, ell, and emm."
            )
        freq_idx, ell, emm = key
        ix = self.getidx(ell, np.abs(emm))
        coeff = self.alm[freq_idx, ix]
        if emm < 0:
            coeff = (-1) ** emm * coeff.conj()
        return coeff

    def all_zero(self):
        self.alm = np.zeros(self.shape, dtype=np.complex128)

    @property
    def shape(self):
        """
        Get the expected shape of the spherical harmonics.
        """
        Nfreq = self.frequencies.size
        shape = (Nfreq, self.size)
        return shape

    @classmethod
    def from_healpix(cls, hp_obj, lmax=None):
        """
        Construct an Alm from a HealpixMap object.
        """
        alm = hp_obj.alm(lmax=lmax)
        if lmax is None:
            lmax = hp.Alm.getlmax(alm.size)
        obj = cls(
            alm=alm,
            lmax=lmax,
            frequencies=hp_obj.frequencies,
            coords=hp_obj.coords,
        )
        return obj

    @classmethod
    def from_angles(
        cls,
        data,
        theta,
        phi,
        frequencies=None,
        lmax=None,
        coords="topographic",
    ):
        """
        Construct an Alm from a grid in theta and phi.
        """
        raise NotImplementedError

    def switch_coords(self, to_coords):
        rotated_alm = rotations.rotate_alm(
            self.alm, from_coords=self.coords, to_coords=to_coords
        )
        self.alm = rotated_alm
        self.coords = to_coords

    def getlm(self, i=None):
        """
        Get the ell and emm corresponding to the numpy index of the alm
        array.
        """
        return super().getlm(self.lmax, i=i)

    def getidx(self, ell, emm):
        """
        Get the index of the alm array for a given ell and emm.
        """
        if not (0 <= emm <= ell <= self.lmax):
            raise ValueError("Ell or emm are out of bounds.")
        return super().getidx(self.lmax, ell, emm)

    @property
    def size(self):
        """
        Get the size of the alm array.
        """
        return super().getsize(self.lmax, mmax=self.lmax)

    @property
    def getlmax(self):
        """
        Get the maxmium ell of the Alm object.
        """
        return self.lmax

    def hp_map(self, nside=64):
        """
        Construct a healpy map from the Alm.
        """
        return alm2map(self.alm, nside=nside, mmax=None)

    def rotate_alm_angle(self, phi):
        """
        Get the coefficients that rotate the alms around the z-axis by phi
        (measured counterclockwise).

        Parameters
        ----------
        phi : array-like
            The angle(s) to rotate the azimuth by in radians.

        Returns
        -------
        phase : np.ndarray
            The coefficients that rotate the alms by phi. Will have shape
            (phi.size, 1, alm.size) to be broadcastable with alm shapes.

        """
        phi = np.ravel(phi)
        phase = rotations.rot_alm_z(phi, self.lmax)
        phase.shape = (phi.size, 1, self.size)
        return phase

    def rotate_alm_time(self, times, world="moon"):
        """
        Rotate alms in time counterclockwise around the z-axis.
        """
        world = world.lower()
        if world == "moon":
            sidereal_day = constants.sidereal_day_moon
        elif world == "earth":
            sidereal_day = constants.sidereal_day_earth
        else:
            raise ValueError(f"World must be 'moon' or 'earth', not {world}.")
        dphi = 2 * np.pi * times / sidereal_day
        phase = self.rotate_alm_angle(dphi)
        return phase



from astropy.coordinates import AltAz, EarthLocation, ICRS
from astropy import units
import healpy as hp
from lunarsky import LunarTopo, MCMF, MoonLocation, Time
import numpy as np
from spiceypy import pxform

from .constants import PIX_WEIGHTS_NSIDE


def topo2radec(theta, phi, time, loc, grid=True):
    """
    Convert topocentric coordinates to ra/dec at given time. Useful for
    antenna beams.

    Parameters
    ----------
    phi : array-like
        The azimuth(s) in radians.
    theta : array-like
        The zenith angle(s) in radians.
    time : array-like, str, or Time instance
        The time to compute the transformation at. Must be able to initialize
        an astropy.time.Time object.
    loc : array-like (lat, lon, alt) or EarthLocation instance
        The location of the topocentric coordinates. Must be able to initialize
        an astropy.coordinates.EarthLocation object.
    grid : bool
        If True then phi and theta are assumed to be coordinate axes of a grid.
        This has no effect if phi and theta have size 1.

    Returns
    -------
    ra : np.ndarray
        The right acension(s) in degrees.
    dec : np.ndarray
        The declination(s) in degrees.

    """
    phi = np.ravel(phi)
    theta = np.ravel(theta)
    if grid:  # phi and theta are coordinate axis
        phi, theta = np.meshgrid(phi, theta)
        phi = phi.ravel()
        theta = theta.ravel()
    # Allow loc to be earth location object and time to be Time object
    if not isinstance(loc, EarthLocation):
        lat, lon, alt = loc
        loc = EarthLocation(
            lat=lat * units.deg, lon=lon * units.deg, height=alt * units.m
        )

    time = Time(time, scale="utc", location=loc)
    azs = phi * units.rad
    alts = (np.pi / 2 - theta) * units.rad
    altaz = AltAz(alt=alts, az=azs, location=loc, obstime=time)
    icrs = altaz.transform_to(ICRS())
    ra = icrs.ra.deg
    dec = icrs.dec.deg
    return ra, dec


def radec2topo(ra, dec, time, loc):
    """
    Convert ra/dec-combination(s) to topocentric coordinates.

    Parameters
    ----------
    ra : array-like
        The right ascension(s) in degrees.
    dec : array-like
        The declination(s) in degrees.
    time : array-like, str, or Time instance
        The time to compute the transformation at. Must be able to initialize
        an astropy.time.Time object.
    loc : array-like (lat, lon, alt) or EarthLocation instance
        The location of the topocentric coordinates. Must be able to initialize
        an astropy.coordinates.EarthLocation object.

    Returns
    -------
    theta : np.ndarray
        Colatitudes in radians.
    phi : np.ndarray
        Azimuths in radians.

    """
    if not isinstance(loc, EarthLocation):
        lat, lon, alt = loc
        loc = EarthLocation(
            lat=lat * units.deg, lon=lon * units.deg, height=alt * units.m
        )

    time = Time(time, scale="utc", location=loc)

    icrs = ICRS(ra=ra * units.deg, dec=dec * units.deg)
    # transform to altaz
    altaz = icrs.transform_to(AltAz(location=loc, obstime=time))
    theta = np.pi / 2 - altaz.alt.rad
    phi = altaz.az.rad
    return theta, phi


def topo2mcmf(theta, phi, time, loc, grid=True):
    """
    Convert topocentric coordinates to mcmf at given time. Useful for
    antenna beams.

    Parameters
    ----------
    phi : array-like
        The azimuth(s) in radians.
    theta : array-like
        The zenith angle(s) in radians.
    time : array-like, str, or Time instance
        The time to compute the transformation at. Must be able to initialize
        an astropy.time.Time object.
    loc : array-like (lat, lon, alt) or EarthLocation instance
        The location of the topocentric coordinates. Must be able to initialize
        an astropy.coordinates.EarthLocation object.
    grid : bool
        If True then phi and theta are assumed to be coordinate axes of a grid.
        This has no effect if phi and theta have size 1.

    Returns
    -------
    lon : np.ndarray
        MCMF longtitude in degrees.
    lat : np.ndarray
        MCMF latitude in degrees.

    """
    phi = np.ravel(phi)
    theta = np.ravel(theta)
    if grid:  # phi and theta are coordinate axis
        phi, theta = np.meshgrid(phi, theta)
        phi = phi.ravel()
        theta = theta.ravel()
    # Allow loc to be earth location object and time to be Time object
    if not isinstance(loc, MoonLocation):
        lat, lon, alt = loc
        loc = MoonLocation(
            lat=lat * units.deg, lon=lon * units.deg, height=alt * units.m
        )

    time = Time(time, scale="utc", location=loc)
    azs = phi * units.rad
    alts = (np.pi / 2 - theta) * units.rad
    altaz = LunarTopo(alt=alts, az=azs, location=loc, obstime=time)
    mcmf = altaz.transform_to(MCMF())
    lon = mcmf.spherical.lon.deg
    lat = mcmf.spherical.lat.deg
    return lon, lat


def mcmf2topo(lon, lat, time, loc):
    """
    Convert ra/dec-combination(s) to topocentric coordinates.

    Parameters
    ----------
    lon : array-like
        Longtitude in degrees.
    lat : array-like
        Latitude in degrees.
    time : array-like, str, or Time instance
        The time to compute the transformation at. Must be able to initialize
        an astropy.time.Time object.
    loc : array-like (lat, lon, alt) or EarthLocation instance
        The location of the topocentric coordinates. Must be able to initialize
        an astropy.coordinates.EarthLocation object.

    Returns
    -------
    theta : np.ndarray
        Colatitudes in radians.
    phi : np.ndarray
        Azimuths in radians.

    """
    if not isinstance(loc, MoonLocation):
        obs_lat, obs_lon, alt = loc
        loc = MoonLocation(
            lat=obs_lat * units.deg,
            lon=obs_lon * units.deg,
            height=alt * units.m,
        )

    time = Time(time, scale="utc", location=loc)

    mcmf = MCMF(
        lon=lon * units.deg,
        lat=lat * units.deg,
        representation_type="spherical",
    )

    # transform to altaz
    altaz = mcmf.transform_to(LunarTopo(location=loc, obstime=time))
    theta = np.pi / 2 - altaz.alt.rad
    phi = altaz.az.rad
    return theta, phi


def rot_coords(
    axis1, axis2, from_coords, to_coords, time=None, loc=None, lonlat=False
):
    """
    Wrapper for the other coordinate transform functions. Rotates coordinates
    from one coordinate system to another.
    Supported coordinate system conversions are
        topocentric <-> mcmf (asssumed topocentric at a MoonLocation)
        topocentric <-> equatorial  (topocenric at an EarthLocation)

    Parameters
    ----------
    axis1 : 1d-array
        Colatitudes in radians (if lonlat = False) or longtitudes in degrees
        (if lonlat = True).

    axis2 : 1d-array
        Azimuth angles in radians (if lonlat = False) or latitudes in degrees
        (if lonlat = True).

    from_coords : str
        Coordinate system to transform from.

    to_coords : str
        Coordinate system to transform to.

    time : str or astropy.time.Time instance or lunarsky.time.Time instance
        The time of the coordinate transform. Must be able to initialize a Time
        object. Required if transforming to or from topocentric coordinates.

    loc: array-like (lat, lon, alt) or astropy.coordinates.EarthLocation
         instance or lunarsky.moon.MoonLocation instance
        The location of the coordinate transform. Required if transforming
        to or from topocentric coordinates. If array-like it must have the
        form (latitude, longtitude, altitude).

    lonlat : bool
        If True, input and out are longtitudes and latitudes in degrees.
        Otherwise, they are colatitudes (polar angle) and azimuths in radians.

    Returns
    -------
    rot_axis1 : np.1darray
        Colatitudes in radians in new coordinate system (if lonlat = False) or
        longtitudes in degrees (in lonlat = True).

    rot_axis2: np.1darray
        Azimuths in radians in new coordinate system (if lonlat = False) or
        latitudes in degrees (if lonlat = True).

    """
    from_coords = from_coords.lower()
    to_coords = to_coords.lower()

    if from_coords == to_coords:
        return axis1, axis2

    elif from_coords == "topocentric":
        if to_coords == "equatorial":
            func = topo2radec
        elif to_coords == "mcmf":
            func = topo2mcmf
        else:
            raise ValueError(
                "Coordinate transform must be topocentric <-> equatorial or "
                f"topocentric <-> mcmf, not {from_coords} to {to_coords}."
            )

        if lonlat:
            phi = np.deg2rad(axis1)
            theta = np.pi / 2 - np.deg2rad(axis2)
        else:
            theta = axis1
            phi = axis2
        ra, dec = func(theta, phi, time, loc, grid=False)
        if lonlat:
            return ra, dec
        else:
            rot_axis1 = np.pi / 2 - np.deg2rad(dec)
            rot_axis2 = np.deg2rad(ra)
            return rot_axis1, rot_axis2

    elif to_coords == "topocentric":
        if from_coords == "equatorial":
            func = radec2topo
        elif from_coords == "mcmf":
            func = mcmf2topo
        else:
            raise ValueError(
                "Coordinate transform must be topocentric <-> equatorial or "
                f"topocentric <-> mcmf, not {from_coords} to {to_coords}."
            )

        if lonlat:
            ra = axis1
            dec = axis2
        else:
            dec = 90 - np.rad2deg(axis1)
            ra = np.rad2deg(axis2)
        theta, phi = func(ra, dec, time, loc)
        if lonlat:
            rot_axis1 = np.rad2deg(phi)
            rot_axis2 = 90 - np.rad2deg(theta)
            return rot_axis1, rot_axis2
        else:
            return theta, phi


def get_euler(from_coords="galactic", to_coords="mcmf", time=None):
    """
    Compute the (ZYX) Euler angles needed to describe a rotation between MCMF,
    galactic and equatorial coordinates.

    Parameters
    ----------
    time : str or astropy.timing.Time instance
        The time of the coordinate transform.

    from_coords : str (optional)
        Coordinate system to convert from.

    to_coords : str (optional)
        Coordinate system to convert to.

    Returns
    -------
    euler_angles : tup
        The ZYX Euler angles in radians that describe the coordinate
        transformation.

    """
    fc = from_coords.lower()
    tc = to_coords.lower()
    coords = {"galactic": "GALACTIC", "equatorial": "J2000", "mcmf": "MOON_ME"}
    if fc not in coords or tc not in coords:
        raise ValueError(
            f"Invalid coordinate system name, must be in {list(coords.keys())}"
        )
    if time is None:
        et = 0
    else:  # get epoch time
        et = Time(time) - Time("J2000", scale="tt")
        et = et.sec
    # rotation matrix
    rot_mat = pxform(coords[fc], coords[tc], et)
    # get euler angles
    beta = -np.arcsin(rot_mat[0, 2])
    alpha = np.arctan2(
        rot_mat[1, 2] / np.cos(beta), rot_mat[2, 2] / np.cos(beta)
    )
    gamma = np.arctan2(
        rot_mat[0, 1] / np.cos(beta), rot_mat[0, 0] / np.cos(beta)
    )
    euler_angles = (gamma, -beta, alpha)
    return euler_angles


def hp_rotate(from_coords, to_coords, time=None):
    """
    Parameters
    ----------
    time : str or astropy.timing.Time instance
        The time of the coordinate transform.
    """
    hp_coords = {"galactic": "G", "equatorial": "C"}
    fc = from_coords.lower()
    tc = to_coords.lower()
    if "mcmf" in [fc, tc]:
        euler_angles = get_euler(from_coords=fc, to_coords=tc, time=time)
        rot = hp.Rotator(rot=euler_angles, deg=False, eulertype="ZYX")
    elif fc in hp_coords and tc in hp_coords:
        rot = hp.Rotator(coord=[hp_coords[fc], hp_coords[tc]])
    else:
        raise ValueError(
            f"Invalid coordinate system, must be in {list(hp_coords.keys())}"
        )
    return rot


def rotate_map(sky_map, from_coords="galactic", to_coords="mcmf"):
    rot = hp_rotate(from_coords, to_coords)
    sky_map = np.array(sky_map, copy=True, dtype=np.float64)
    npix = sky_map.shape[-1]
    nside = hp.npix2nside(npix)
    use_pix_weights = nside in PIX_WEIGHTS_NSIDE
    if sky_map.ndim == 1:
        rotated_map = rot.rotate_map_alms(
            sky_map, use_pixel_weights=use_pix_weights
        )
    elif sky_map.ndim == 2:
        rotated_map = np.empty_like(sky_map)
        for i, m in enumerate(sky_map):  # each frequency
            rm = rotate_map(m, from_coords=from_coords, to_coords=to_coords)
            rotated_map[i] = rm
    else:
        raise ValueError("sky_map must be a 1d map or a (2d) list of maps.")
    return rotated_map


def rotate_alm(alm, from_coords="galactic", to_coords="mcmf"):
    rot = hp_rotate(from_coords, to_coords)
    alm = np.array(alm, copy=True, dtype=np.complex128)
    if alm.ndim == 1:
        rotated_alm = rot.rotate_alm(alm)
    elif alm.ndim == 2:
        rotated_alm = np.empty_like(alm)
        for i, a in enumerate(alm):  # for each frequency
            rotated_alm[i] = rotate_alm(
                a, from_coords=from_coords, to_coords=to_coords
            )
    else:
        raise ValueError(f"alm must have 1 or 2 dimensions, not {alm.ndim}.")
    return rotated_alm


def rot_alm_z(phi, lmax):
    """
    Get the coefficients that rotate alms around the z-axis by phi
    (measured counterclockwise).

    Parameters
    ----------
    phi : array-like
        The angle(s) to rotate the azimuth by in radians.
    lmax : int
        The maximum ell of the alm.

    Returns
    -------
     phase : np.ndarray
        The coefficients that rotate the alms by phi. Will have shape
        (alm.size) if phi is a scalar or (phi.size, alm.size) if phi
        is an array.

    """

    phi = np.reshape(phi, (-1, 1))
    emms = hp.Alm.getlm(lmax)[1].reshape(1, -1)
    phase = np.exp(1j * emms * phi)
    return np.squeeze(phase)



import healpy as hp
import numpy as np
from pygdsm import GlobalSkyModel2016 as GSM
from .healpix import HealpixMap


class Sky(HealpixMap):
    def __init__(
        self,
        sky_map=None,
        frequencies=None,
        nested_input=False,
        coords="galactic",
    ):
        """
        Class that holds Sky objects. Thin wrapper for HealpixMap objects.
        """
        super().__init__(
            data=sky_map,
            nside=None,
            nested_input=nested_input,
            frequencies=frequencies,
            coords=coords,
        )

    @classmethod
    def from_alm(cls, alm_obj, nside=None):
        """
        Construct a Sky instance from an Alm object (defined below).
        """
        if nside is None:
            nside = (alm_obj.lmax + 1) // 3
        hp_map = alm_obj.hp_map(nside=nside)
        obj = cls(
            sky_map=hp_map,
            frequencies=alm_obj.frequencies,
            nested_input=False,
            coords=alm_obj.coords,
        )
        return obj

    @classmethod
    def gsm(
        cls,
        freq,
        power_law=False,
        gen_freq=None,
        spectral_index=None,
        nside=None,
    ):
        """
        Construct a sky object with pygdsm.

        Parameters
        ----------
        freq : array-like
            Frequencies to make map at.
        power_law : bool
            Make maps at many frequencies assuming the frequency dependence is
            a power law.
        gen_freq : float
            Which frequency to create the sky map at with GSM. Only matters if
            ``power_law''=True, in which case this map will be extrapolated to
            ``freq''.
        spectral_index : float
            Spectral index of power law. Only applies if ``power_law''=True.
        nside : int
            Set the nside parameter of the resulting sky map, i.e., increasing
            or decreasing the number of pixels in the map. Note that this does
            not increase the information content in the map, just splits pixels
            to smaller pixels or average them to fewer pixels.
        """
        freq = np.ravel(freq)
        # frequencies to generate maps at
        if not power_law:
            gen_freq = freq  # generate at all freqs
        elif gen_freq is None:
            # if power law, only generate one map with GSM
            # use largest frequency since it's lkely inside the GSM range
            gen_freq = freq[-1]

        gsm16 = GSM(freq_unit="MHz", data_unit="TRJ", resolution="lo")
        sky_map = gsm16.generate(gen_freq)

        if power_law:
            kwargs = {
                "ref_map": sky_map,
                "ref_freq": gen_freq,
                "return_map": True,
            }
            if spectral_index is not None:
                kwargs["spectral_index"] = spectral_index
            sky_map = Sky().power_law_map(freq, **kwargs)

        if nside:
            sky_map = hp.ud_grade(sky_map, nside)

        obj = cls(
            sky_map,
            frequencies=freq,
            nested_input=False,
            coords="galactic",
        )

        return obj

    def power_law_map(
        self,
        freq_out,
        spectral_index=-2.5,
        ref_map=None,
        ref_freq=None,
        return_map=False,
    ):
        """
        Extrapolate or interpolate a map at specified frequencies to other
        frequencies using a power law of specified spectral index.

        Parameters
        ----------
        freq_out : array-like
            The frequencies to extrapolate the map to.
        spectral_index : float
            The spectral index of the power law. Defaults to -2.5
            (theoretical synchrotron power law).
        ref_map : np.ndarray (optional)
            The reference map to extrapolate. If None, use self.data.
        ref_freq : array-like (optional)
            The reference frequency that the map is given at. If None, use
            self.frequencies.
        return_map : bool
            Set to True to return the map and frequencies as np.ndarrays.
            Otherwise, the map and frequencies replace the attributes sky_map
            and the frequencies.

        Returns
        -------
        These will only be returned if the flag return_map is True.

        sky_map : np.ndarray
            The extrapolated healpix map at the given frequencies.
        freq_out : array-like
           The frequencies of the extrapolated map.

        Raises
        ------
        ValueError :
            If the reference maps or frequencies are None, their
            shapes don't match.
            If multiple maps are provided but they don't follow the power law
            specified.

        """
        if ref_freq is None:
            ref_freq = self.frequencies
        else:
            ref_freq = np.ravel(ref_freq).copy()

        if ref_map is None:
            ref_map = self.data
        else:
            ref_map = np.array(ref_map, copy=True)
            if ref_map.ndim == 1:
                ref_map.shape = (1, -1)

        def _pow_law(f):
            t0 = ref_map[0].reshape(1, -1)
            f0 = ref_freq[0]
            f = np.array(f).reshape(-1, 1)
            return t0 * (f / f0) ** spectral_index

        if not np.allclose(_pow_law(ref_freq), ref_map):
            raise ValueError(
                "The provided maps don't satsify the specified power law."
            )

        sky_map = _pow_law(freq_out)

        if return_map:
            return sky_map

        else:
            self.data = sky_map
            self.frequencies = np.ravel(freq_out)



from copy import deepcopy
import healpy
import numpy as np
import pytest
from croissant import rotations, healpix as hp
from croissant.constants import sidereal_day_earth, sidereal_day_moon, Y00


def test_healpix2lonlat():
    nside = 32
    pix = 1  # the north pole
    lon, lat = hp.healpix2lonlat(nside, pix=pix)
    assert lat > 85  # should be close to 90

    lons, lats = hp.healpix2lonlat(nside, pix=None)
    assert lons.shape == lats.shape == (healpy.nside2npix(nside),)
    assert 0 <= lons.all() < 360
    assert -90 <= lats.all() < 90


def test_grid_interp():
    # create mock data at 1 deg spacing in theta/phi
    theta = np.linspace(0, np.pi, 181)
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    data = np.cos(theta) ** 2
    data = np.repeat(data.reshape(-1, 1), phi.size, axis=1)

    # test that we get the same thing back if interpolate to same values
    # interpolation must be called on the flattened arrays
    to_phi, to_theta = [d.ravel() for d in np.meshgrid(phi, theta)]
    interp_data = hp.grid_interp(data, theta, phi, to_theta, to_phi)
    # interp data has a frequency axis as the 0th axis
    assert interp_data.shape == (1, to_theta.size)
    assert np.allclose(data.ravel(), interp_data[0])

    # test with data that depends on phi
    data = np.repeat(phi.reshape(1, -1), theta.size, axis=0)
    interp_data = hp.grid_interp(data, theta, phi, to_theta, to_phi)
    assert interp_data.shape == (1, to_theta.size)
    # we need to disregard the poles here
    no_poles = (0 < to_theta) & (to_theta < np.pi)
    assert np.allclose(data.ravel()[no_poles], interp_data[0, no_poles])


def test_grid2healpix():
    # constant map
    theta = np.linspace(0, np.pi, 181)
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    data = np.ones((theta.size, phi.size))
    nside = 32
    npix = healpy.nside2npix(nside)
    hp_map = hp.grid2healpix(data, nside, theta=theta, phi=phi)
    expected_map = np.ones((1, npix))
    assert np.allclose(hp_map, expected_map)

    # map that depends on theta
    phi_grid, theta_grid = np.meshgrid(phi, theta)
    data = np.sin(theta_grid)
    hp_map = hp.grid2healpix(data, nside, theta=theta, phi=phi)
    # angles of healpix map:
    lat = hp.healpix2lonlat(nside)[1]
    hp_theta = np.pi / 2 - np.deg2rad(lat)
    expected_map = np.sin(hp_theta)
    assert np.allclose(hp_map, expected_map)

    # map that depends on phi
    phi_grid, theta_grid = np.meshgrid(phi, theta)
    data = np.sin(phi_grid)
    hp_map = hp.grid2healpix(data, nside, theta=theta, phi=phi)
    # angles of healpix map:
    lon = hp.healpix2lonlat(nside)[0]
    hp_phi = np.deg2rad(lon)
    expected_map = np.sin(hp_phi)
    # for map that depends on phi we expect errors at the pole
    pole_th = 0.15  # radians, where we allow errors (theta)
    pix = healpy.ang2pix(nside, pole_th, 0)
    assert np.allclose(hp_map[0, pix:-pix], expected_map[pix:-pix])

    # set pixel centers to be the first npix values of theta/phi
    data = np.sin(theta_grid)
    pc_theta = theta_grid.ravel()[:npix]
    pc_phi = phi_grid.ravel()[:npix]
    pix_centers = np.transpose([pc_theta, pc_phi])
    hp_map = hp.grid2healpix(
        data, nside, theta=theta, phi=phi, pixel_centers=pix_centers
    )
    expected_map = np.sin(pc_theta)
    assert np.allclose(hp_map, expected_map)


def test_map2alm():
    nside = 32
    npix = healpy.nside2npix(nside)
    data = np.ones(npix)
    lmax = 5
    alm = hp.map2alm(data, lmax)
    assert np.isclose(alm[0], Y00 * 4 * np.pi)  # a00 = Y00 * 4pi
    assert np.allclose(alm[1:], 0)  # rest of alms should be 0

    # multiple maps at once
    c = 10  # constant extra value
    d0 = np.arange(npix)
    d1 = d0 + c
    alm0 = hp.map2alm(d0, lmax)
    alm1 = hp.map2alm(d1, lmax)
    data = np.array([d0, d1])
    alm = hp.map2alm(data, lmax)
    # check that computing jointly is the same as individually
    assert np.allclose(alm[0], alm0)
    assert np.allclose(alm[1], alm1)
    # d1 is d0 + constant value so they should only differ at a00
    assert np.allclose(alm0[1:], alm1[1:])
    assert np.isclose(alm0[0] + c * Y00 * 4 * np.pi, alm1[0])


def test_alm2map():
    # make constant map with mmax = lmax
    lmax = 10
    mmax = lmax
    a00 = 3
    size = healpy.Alm.getsize(lmax, mmax=mmax)
    alm = np.zeros(size, dtype=np.complex128)
    alm[0] = a00
    nside = 32
    hp_map = hp.alm2map(alm, nside=nside, mmax=mmax)
    npix = hp_map.shape[-1]
    assert nside == healpy.npix2nside(npix)
    expected_map = np.full(npix, a00 * Y00)
    assert np.allclose(hp_map, expected_map)

    # constant map with mmax < lmax
    mmax = 5
    size = healpy.Alm.getsize(lmax, mmax=mmax)
    alm = np.zeros(size, dtype=np.complex128)
    alm[0] = a00
    hp_map = hp.alm2map(alm, nside=nside, mmax=mmax)
    npix = hp_map.shape[-1]
    assert nside == healpy.npix2nside(npix)
    expected_map = np.full(npix, a00 * Y00)
    assert np.allclose(hp_map, expected_map)

    # make many maps
    frequencies = np.linspace(1, 50, 50)
    mmax = lmax
    size = healpy.Alm.getsize(lmax, mmax=mmax)
    alm = np.zeros((frequencies.size, size), dtype=np.complex128)
    alm[:, 0] = a00 * frequencies**2.5
    hp_map = hp.alm2map(alm, nside=nside, mmax=mmax)
    expected_maps = np.full((frequencies.size, npix), a00 * Y00)
    expected_maps *= frequencies.reshape(-1, 1) ** 2.5
    assert np.allclose(hp_map, expected_maps)

    # inverting map2alm
    lmax = 10
    mmax = lmax
    size = healpy.Alm.getsize(lmax, mmax=mmax)
    alm = np.zeros(size, dtype=np.complex128)
    lm_dict = {
        (0, 0): 5.1,
        (2, 0): 7.4,
        (3, 2): 3 + 5j,
        (4, 1): -4.3 + 1.3j,
        (7, 7): 11,
    }
    for ell, emm in lm_dict:
        ix = healpy.Alm.getidx(lmax, ell, emm)
        alm[ix] = lm_dict[(ell, emm)]

    nside = 64
    m = hp.alm2map(alm, nside=nside)
    alm_ = hp.map2alm(m, lmax)
    assert np.allclose(alm, alm_)
    m_ = hp.alm2map(alm_, nside=nside)
    assert np.allclose(m, m_)


def test_nested_input():
    freqs = np.linspace(1, 50, 50)
    nside = 10  # this is invalid for NESTED but OK for RING
    npix = healpy.nside2npix(nside)
    data = 10 * np.ones((freqs.size, npix))
    kwargs = {
        "data": data,
        "nside": nside,
        "nested_input": False,
        "frequencies": freqs,
    }
    hp.HealpixMap(**kwargs)  # should work
    # should raise an error if nested is True
    kwargs["nested_input"] = True
    with pytest.raises(ValueError):
        hp.HealpixMap(**kwargs)

    # valid nested input
    nside = 8
    npix = healpy.nside2npix(nside)
    data = np.arange(npix) ** 2
    kwargs = {
        "data": data,
        "nside": nside,
        "nested_input": True,
        "frequencies": None,
    }
    hp_map = hp.HealpixMap(**kwargs)
    # hp_map data is in the RING scheme:
    ring_ix = healpy.nest2ring(nside, np.arange(npix))
    data_ring = data[ring_ix]
    assert np.allclose(hp_map.data, data_ring)


def test_from_alm():
    nside = 10
    lmax = 3 * nside - 1
    alm = hp.Alm(lmax=lmax)
    a00 = 10
    alm[0, 0] = a00
    hp_map = hp.HealpixMap.from_alm(alm, nside=None)
    # healpix map should be able to infer the nside from lmax
    assert hp_map.nside == nside
    npix = healpy.nside2npix(nside)
    # the map should just be = a00 * Y00 everywhere
    expected_map = np.full((1, npix), a00 * Y00)
    assert np.allclose(hp_map.data, expected_map)


def test_ud_grade():
    freqs = np.linspace(1, 50, 50)
    nside = 8
    npix = healpy.nside2npix(nside)
    data = np.repeat(np.arange(npix).reshape(1, -1), freqs.size, axis=0)
    kwargs = {
        "data": data,
        "nside": nside,
        "nested_input": False,
        "frequencies": freqs,
    }
    hpm = hp.HealpixMap(**kwargs)
    nside_out = [1, 2, 8, 32]
    for ns in nside_out:
        hp_copy = deepcopy(hpm)
        hp_copy.ud_grade(ns)
        assert hp_copy.nside == ns
        assert np.allclose(hp_copy.data, healpy.ud_grade(hpm.data, ns))


def test_switch_coords():
    nside = 8
    npix = healpy.nside2npix(nside)
    data = np.arange(npix)
    # switch from galactic to equatorial
    coords = "galactic"
    new_coords = "equatorial"
    hp_map = hp.HealpixMap(data=data, nside=nside, coords=coords)
    assert hp_map.coords == coords
    hp_map.switch_coords(new_coords)
    assert hp_map.coords == new_coords
    expected_data = rotations.rotate_map(
        data, from_coords=coords, to_coords=new_coords
    )
    assert np.allclose(hp_map.data, expected_data)

    # several maps at once
    freqs = np.arange(10).reshape(-1, 1)
    data = np.arange(npix).reshape(1, -1) * freqs
    hp_map = hp.HealpixMap(
        data=data, nside=nside, frequencies=freqs, coords=coords
    )
    assert hp_map.coords == coords
    hp_map.switch_coords(new_coords)
    assert hp_map.coords == new_coords
    expected_data = rotations.rotate_map(
        data, from_coords=coords, to_coords=new_coords
    )
    assert np.allclose(hp_map.data, expected_data)


def test_alm():
    nside = 32
    npix = healpy.nside2npix(nside)
    data = np.arange(npix)
    hp_map = hp.HealpixMap(data=data, nside=nside)
    # test default lmax
    alm = hp_map.alm(lmax=None)
    expected_lmax = 3 * nside - 1  # should be default
    expected_size = healpy.Alm.getsize(expected_lmax, mmax=expected_lmax)
    assert alm.shape == (1, expected_size)
    # specify lmax
    lmax = 10
    alm = hp_map.alm(lmax=lmax)
    expected_size = healpy.Alm.getsize(lmax, mmax=lmax)
    assert alm.shape == (1, expected_size)

    # several maps at once
    freqs = np.arange(10).reshape(-1, 1)
    data = np.arange(npix).reshape(1, -1) * freqs
    hp_map = hp.HealpixMap(data=data, nside=nside, frequencies=freqs)
    alm = hp_map.alm(lmax=lmax)
    assert alm.shape == (freqs.size, expected_size)


def test_Alm_init():
    lmax = 5
    size = healpy.Alm.getsize(lmax, mmax=lmax)
    freqs = np.linspace(1, 50, 50)
    # alm = None
    alm = hp.Alm(alm=None, lmax=lmax, frequencies=freqs)
    expected_alm = np.zeros((freqs.size, size), dtype=np.complex128)
    assert np.allclose(alm.alm, expected_alm)
    # lmax = None
    alms = np.arange(size).reshape(1, -1) * freqs.reshape(-1, 1)
    alm = hp.Alm(alm=alms, lmax=None, frequencies=freqs)
    assert alm.lmax == lmax  # init should infer lmax from alm size
    # both alm and lmax are specfied but inconsistent
    wrong_lmax = 10
    with pytest.raises(ValueError):
        hp.Alm(alm=alms, lmax=wrong_lmax, frequencies=freqs)


def test_alm_indexing():
    lmax = 10
    freqs = np.linspace(1, 50, 50)
    nfreqs = freqs.size
    # initialize all alms to 0
    alm = hp.Alm(alm=None, lmax=lmax, frequencies=freqs)
    # set a00 = 1 for first half of frequencies
    alm[: nfreqs // 2, 0, 0] = 1.0
    # check __setitem__ acted correctly on alm.alm
    assert np.allclose(alm.alm[: nfreqs // 2, 0], 1)
    assert np.allclose(alm.alm[nfreqs // 2 :, 0], 0)
    assert np.allclose(alm.alm[:, 1:], 0)
    # check that __getitem__ agrees:
    assert np.allclose(alm[: nfreqs // 2, 0, 0], 1)
    assert np.allclose(alm[nfreqs // 2 :, 0, 0], 0)
    # __getitem__ can't get multiple l-modes or m-modes at once...
    for ell in range(1, lmax + 1):
        for emm in range(ell + 1):
            assert np.allclose(alm[:, ell, emm], 0)

    # set everything back to 0
    alm.all_zero()
    # negative indexing
    val = 3.0 + 2.3j
    alm[-1, 10, 7] = val
    assert alm[-1, 10, 7] == val
    ix = healpy.Alm.getidx(lmax, 10, 7)
    assert alm[-1, 10, 7] == alm.alm[-1, ix]

    # negative emm
    alm.all_zero()
    alm[0, 3, 2] = val
    assert np.isclose(alm[0, 3, -2], (-1) ** 2 * np.conj(val))

    # frequency index not specified
    with pytest.raises(IndexError):
        alm[3, 2] = 5
        alm[7, -1]

    # no frequencies
    alm = hp.Alm(alm=None, lmax=lmax, frequencies=None)
    alm[5, 2] = 3.0
    assert alm[5, 2] == 3.0
    assert alm[0, 5, 2] == 3.0  # can optionally have freq idx = 0
    assert alm[-1, 5, 2] == 3.0  # ... or -1


def test_from_healpix():
    nside = 8
    npix = healpy.nside2npix(nside)
    freqs = np.linspace(1, 50, 50)
    data = np.arange(npix).reshape(1, -1) * freqs.reshape(-1, 1) ** 2
    coords = "equatorial"
    hp_map = hp.HealpixMap(
        data=data, nside=nside, frequencies=freqs, coords=coords
    )
    lmax = 10
    alm = hp.Alm.from_healpix(hp_map, lmax=lmax)
    assert alm.lmax == lmax
    assert np.allclose(alm.frequencies, freqs)
    assert alm.coords == coords
    assert np.allclose(alm.alm, hp.map2alm(data, lmax))


def test_alm_switch_coords():
    lmax = 10
    size = healpy.Alm.getsize(lmax, mmax=lmax)
    data = np.arange(size)
    # switch from galactic to equatorial
    coords = "galactic"
    new_coords = "equatorial"
    alm = hp.Alm(alm=data, lmax=lmax, coords=coords)
    assert alm.coords == coords
    alm.switch_coords(new_coords)
    assert alm.coords == new_coords
    expected_data = rotations.rotate_alm(
        data, from_coords=coords, to_coords=new_coords
    )
    assert np.allclose(alm.alm, expected_data)

    # several alms at once
    freqs = np.arange(10).reshape(-1, 1)
    data = np.arange(size).reshape(1, -1) * freqs
    alm = hp.Alm(alm=data, lmax=lmax, frequencies=freqs, coords=coords)
    assert alm.coords == coords
    alm.switch_coords(new_coords)
    assert alm.coords == new_coords
    expected_data = rotations.rotate_alm(
        data, from_coords=coords, to_coords=new_coords
    )
    assert np.allclose(alm.alm, expected_data)


def test_getidx():
    lmax = 5
    alm = hp.Alm(lmax=lmax)
    ell = 3
    emm = 2
    bad_ell = 2 * lmax  # bigger than lmax
    bad_emm = 4  # bigger than ell
    with pytest.raises(ValueError):
        alm.getidx(bad_ell, emm)
        alm.getidx(ell, bad_emm)
        alm.getidx(-ell, emm)  # should fail since l < 0
        alm.getidx(ell, -emm)  # shoul fail since m < 0

    # try convert back and forth ell, emm <-> index
    ix = alm.getidx(ell, emm)
    ell_, emm_ = alm.getlm(i=ix)
    assert ell == ell_
    assert emm == emm_


def test_hp_map():
    # make constant map
    lmax = 10
    nside = 100
    alm = hp.Alm(lmax=lmax)
    a00 = 5
    alm[0, 0] = a00
    hp_map = alm.hp_map(nside=nside)
    npix = hp_map.shape[-1]
    assert nside == healpy.npix2nside(npix)
    expected_map = np.full((1, npix), a00 * Y00)
    assert np.allclose(hp_map, expected_map)

    # make many maps
    frequencies = np.linspace(1, 50, 50)
    alm = hp.Alm(lmax=lmax, frequencies=frequencies)
    alm[:, 0, 0] = a00 * frequencies
    hp_map = alm.hp_map(nside=nside)
    expected_maps = np.full((frequencies.size, npix), a00 * Y00)
    expected_maps *= frequencies.reshape(-1, 1)
    assert np.allclose(hp_map, expected_maps)


def test_rotate_alm_angle():
    lmax = 10
    alm = hp.Alm(lmax=lmax)
    phi = np.pi / 2
    phase = alm.rotate_alm_angle(phi)
    for ell in range(lmax + 1):
        for emm in range(ell + 1):
            ix = alm.getidx(ell, emm)
            assert np.isclose(phase[0, 0, ix], np.exp(1j * emm * phi))

    # rotate a set of angles
    phi = np.linspace(0, 2 * np.pi, num=361)  # 1 deg spacing
    phase = alm.rotate_alm_angle(phi)
    for ell in range(lmax + 1):
        for emm in range(ell + 1):
            ix = alm.getidx(ell, emm)
            assert np.allclose(phase[:, 0, ix], np.exp(1j * emm * phi))

    # check that phi = 0 and phi = 2pi give the same answer
    assert np.allclose(phase[0], phase[-1])


def test_rotate_alm_time():
    alm = hp.Alm(lmax=20, frequencies=np.linspace(1, 50, 50))
    div = [1, 2, 4, 8]
    for d in div:
        dphi = 2 * np.pi / d
        # earth
        dt = sidereal_day_earth / d
        assert np.allclose(
            alm.rotate_alm_time(dt, world="earth"), alm.rotate_alm_angle(dphi)
        )
        # moon
        dt = sidereal_day_moon / d
        assert np.allclose(
            alm.rotate_alm_time(dt, world="moon"), alm.rotate_alm_angle(dphi)
        )



from croissant import dpss
import numpy as np
import pytest


def test_dpss_op():

    x = np.linspace(1, 50, 50)  # target frequencies
    with pytest.raises(ValueError):  # didn't specify any kwargs
        _ = dpss.dpss_op(x)

    with pytest.raises(ValueError):
        _ = dpss.dpss_op(
            x, eigenval_cutoff=1e-12, nterms=10
        )  # too many kwargs

    nterms = 10
    design_matrix = dpss.dpss_op(x, nterms=nterms)
    assert design_matrix.shape == (x.size, nterms)
    assert np.allclose(design_matrix, design_matrix.real)


def test_freq2dpss2freq():
    # generate 2 sets of frequencies that should share dpss modes
    # croissant does this with beam and simulation frequencies
    f1 = np.linspace(10, 60, 50)
    f2 = np.linspace(1, 50, 50)

    x = np.unique(np.concatenate((f1, f2)))
    nterms = 10
    A = dpss.dpss_op(x, nterms=nterms)  # design matrix

    # mock data
    data = [np.array([f**2, np.sin(f)]) for f in [f1, f2]]

    # compute dpss coeffs
    d1 = dpss.freq2dpss(data[0].T, f1, x, A)
    d2 = dpss.freq2dpss(data[1].T, f2, x, A)

    assert d1.shape == d2.shape == (nterms, 2)

    d1_ = dpss.dpss2freq(d1, A)
    assert data[0].T.shape == d1_[np.isin(x, f1)].shape
    d2_ = dpss.dpss2freq(d2, A)
    assert data[1].T.shape == d2_[np.isin(x, f2)].shape



from astropy import units
from copy import deepcopy
import healpy
from lunarsky import Time
import numpy as np
import pytest

from croissant.beam import Beam
from croissant import dpss
from croissant.constants import sidereal_day_earth
from croissant.rotations import radec2topo, rotate_alm
from croissant.healpix import Alm, alm2map, grid_interp
from croissant.simulator import Simulator
from croissant.sky import Sky


# define default params for simulator

frequencies = np.linspace(1, 50, 50).reshape(-1, 1, 1)
theta = np.linspace(0, np.pi, 181).reshape(1, -1, 1)
phi = np.linspace(0, 2 * np.pi, 360, endpoint=False).reshape(1, 1, -1)
power = frequencies**2 * np.cos(theta) ** 2  # dipole
power = np.repeat(power, phi.size, axis=2)
beam = Beam(power, theta=theta, phi=phi, frequencies=frequencies)
sky = Sky.gsm(frequencies, power_law=True, gen_freq=25, spectral_index=-2.5)
loc = (40.0, 137.0, 0.0)
t_start = "2022-06-10 12:59:00"
N_times = 250
delta_t = 3600 * units.s
lmax = 32


def test_simulator_init():
    # check that the times are set conisstently regardless of
    # which parameters that specify it
    delta_t, step = np.linspace(0, sidereal_day_earth, N_times, retstep=True)
    step = step * units.s
    t_end = Time(t_start) + delta_t[-1] * units.s
    # specify end, ntimes:
    sim = Simulator(
        beam, sky, loc, t_start, t_end=t_end, N_times=N_times, lmax=lmax
    )
    assert np.allclose(delta_t, sim.dt)
    assert np.isclose(N_times, sim.N_times)
    # specify end, delta t
    sim = Simulator(
        beam, sky, loc, t_start, t_end=t_end, delta_t=step, lmax=lmax
    )
    assert np.allclose(delta_t, sim.dt)
    assert np.isclose(N_times, sim.N_times)
    # specify ntimes, delta t
    sim = Simulator(
        beam, sky, loc, t_start, N_times=N_times, delta_t=step, lmax=lmax
    )
    assert np.allclose(delta_t, sim.dt)
    assert np.isclose(N_times, sim.N_times)

    # check that power is computed correctly before horizon cut
    assert np.allclose(sim.beam.total_power, beam.total_power)
    # check that the init does the horizon cut
    beam_copy = deepcopy(beam)
    beam_copy.horizon_cut()
    assert np.allclose(sim.beam.data, beam_copy.data)

    # check that the simulation coords are set properly
    assert sim.sim_coords == "mcmf"
    # check sky is in the desired simulation coords
    assert sim.sky.coords == sim.sim_coords
    sky_alm = rotate_alm(sky.alm(lmax=lmax))
    assert np.allclose(sim.sky.alm, sky_alm)

    # check that init works correcttly on earth
    sim = Simulator(
        beam,
        sky,
        loc,
        t_start,
        moon=False,
        N_times=N_times,
        delta_t=step,
        lmax=lmax,
    )
    assert sim.sim_coords == "equatorial"

    # check that we get a UserWarning if delta t does not have units
    delta_t = 2
    with pytest.warns(UserWarning):
        Simulator(
            beam, sky, loc, t_start, N_times=2, delta_t=delta_t, lmax=lmax
        )


def test_beam_alm():
    sim = Simulator(
        beam,
        sky,
        loc,
        t_start,
        moon=False,
        N_times=N_times,
        delta_t=delta_t,
        lmax=lmax,
    )

    # check that input args are set correctly
    assert sim.beam.coords == sim.sim_coords == "equatorial"

    # test beam alm by inverting it
    nside = 128
    beam_map = alm2map(sim.beam.alm, nside=nside)  # in healpix ra/dec
    pix = np.arange(healpy.nside2npix(nside))
    ra, dec = healpy.pix2ang(nside, pix, nest=False, lonlat=True)
    inv_theta, inv_phi = radec2topo(ra, dec, t_start, loc)
    interp_sim_beam = grid_interp(
        sim.beam.data, theta, phi, inv_theta, inv_phi
    )
    diff = beam_map - interp_sim_beam
    rms = np.sqrt(np.mean(diff**2, axis=1)) / np.mean(
        interp_sim_beam, axis=1
    )
    assert rms.max() < 1e-3


def test_compute_dpss():
    sim = Simulator(
        beam, sky, loc, t_start, N_times=N_times, delta_t=delta_t, lmax=lmax
    )
    sim.compute_dpss(nterms=10)
    design_matrix = dpss.dpss_op(frequencies, nterms=10)
    assert np.allclose(design_matrix, sim.design_matrix)
    beam_coeff = dpss.freq2dpss(
        sim.beam.alm, frequencies, frequencies, design_matrix
    )
    assert np.allclose(beam_coeff, sim.beam.coeffs)


def test_run():
    # retrieve constant temperature (spatially) sky with dipole beam
    # sky has power law of index -2.5, beam of index 2
    sky = Sky()
    nside = 32
    npix = healpy.nside2npix(nside)
    frequencies = np.linspace(1, 50, 50)
    sky.power_law_map(
        frequencies,
        spectral_index=-2.5,
        ref_map=np.ones(npix) * 1e7,
        ref_freq=frequencies[0],
    )
    theta = np.linspace(0, np.pi, 181)
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    theta, frequencies, phi = np.meshgrid(theta, frequencies, phi, sparse=True)
    power = frequencies**2 * np.cos(theta) ** 2  # dipole
    power = np.repeat(power, phi.size, axis=2)
    beam = Beam(power, theta=theta, phi=phi, frequencies=frequencies)
    sim = Simulator(
        beam, sky, loc, t_start, N_times=N_times, delta_t=delta_t, lmax=lmax
    )
    sim.run(dpss=False)
    beam_a00 = sim.beam.alm[0, 0]  # a00 @ freq = 1 MHz
    sky_a00 = sim.sky.alm[0, 0]  # a00 @ freq = 1 MHz
    # resulting visibility spectrum should go as nu^-.5
    expected_vis = beam_a00 * sky_a00 * np.squeeze(frequencies) ** (-0.5)
    expected_vis /= sim.beam.total_power
    expected_vis.shape = (1, -1)  # add time axis
    assert np.allclose(sim.waterfall, np.repeat(expected_vis, N_times, axis=0))
    # with dpss
    sim.run(dpss=True, nterms=50)
    assert np.allclose(sim.waterfall, np.repeat(expected_vis, N_times, axis=0))

    # test with nonzero m-modes
    sky_alm = Alm(lmax=lmax, coords="mcmf")
    sky_alm[0, 0] = 1e7
    sky_alm[2, 0] = 1e4
    sky_alm[3, 1] = -20.2 + 20.4j
    sky_alm[6, 6] = 1.0 - 3.0j
    sky = Sky.from_alm(sky_alm, nside=128)

    beam00 = 10  # a00
    beam20 = 5  # a20
    beam31 = 1 + 2j  # a31
    beam66 = -1 - 1.34j  # a66
    beam_alm = np.zeros((1, healpy.Alm.getsize(lmax)), dtype=np.complex128)
    beam_alm[0, 0] = beam00
    beam_alm[0, 2] = beam20
    ix31 = healpy.Alm.getidx(lmax, 3, 1)
    beam_alm[0, ix31] = beam31
    ix66 = healpy.Alm.getidx(lmax, 6, 6)
    beam_alm[0, ix66] = beam66
    beam = Beam(beam_alm, alm=True, coords="mcmf")

    sim = Simulator(
        beam,
        sky,
        loc,
        t_start,
        N_times=1,
        delta_t=delta_t,
        lmax=lmax,
    )

    sim.run(dpss=False)
    expected_vis = (
        sky_alm[0, 0] * beam00
        + sky_alm[2, 0] * beam20
        + 2 * np.real(sky_alm[3, 1] * np.conj(beam31))
        + 2 * np.real(sky_alm[6, 6] * np.conj(beam66))
    )
    expected_vis /= sim.beam.total_power
    assert np.isclose(sim.waterfall, expected_vis)

    # test the einsum computation in dpss mode
    frequencies = np.linspace(1, 50, 50).reshape(-1, 1)
    beam_alm = beam_alm.reshape(1, -1) * frequencies**2
    beam = Beam(beam_alm, frequencies=frequencies, alm=True, coords="mcmf")
    sky = Sky.from_alm(sky_alm, nside=128)
    sky.power_law_map(beam.frequencies, ref_freq=25)
    sim = Simulator(
        beam,
        sky,
        loc,
        t_start,
        N_times=1,
        delta_t=delta_t,
        lmax=lmax,
    )
    sim.run(dpss=True, nterms=10)
    # expected output is dot product of alms in frequency space:
    sky_alm = sim.sky.alm
    beam_alm = sim.design_matrix @ sim.beam.coeffs
    temp_vector = np.empty(frequencies.size)
    for i in range(frequencies.size):
        t = sky_alm[i, : lmax + 1].real.dot(beam_alm[i, : lmax + 1].real)
        t += 2 * np.real(
            sky_alm[i, lmax + 1 :].dot(beam_alm[i, lmax + 1 :].conj())
        )
        temp_vector[i] = t
    # output of simulator
    wfall = sim.waterfall * sim.beam.total_power.reshape(1, -1)
    assert np.allclose(temp_vector, wfall)



from hera_filters.dspec import dpss_operator, fit_solution_matrix
import numpy as np


def dpss_op(
    freq_out,
    filter_half_width=50e-9,
    eigenval_cutoff=None,
    edge_suppression=None,
    nterms=None,
    avg_suppression=None,
):
    """
    Compute operator for transforming data from frequency space to a basis of
    Discrete Prolate Spheroidal Sequences (DPSS). Thin wrapper around
    hera_filters.dspec.dpss_operator suitable for a single antenna
    autocorrelation experiment.

    Specify one of eigenval_cutoff (min eigenvalue of sinc matrix),
    edge_suppression (degree of tone suppression at edge of filter window),
    nterms (number of DPSS terms to include), and
    avg_suppression (the avg suppression of tones within filter edges) to
    determine the cutoff for number of DPSS modes to use.

    See Ewall-Wice et al (2021) and https://github.com/HERA-Team/hera_filters
    for discussion of DPSS.

    Parameters
    ----------
    freq_out : array-like, optional
        The frequencies in MHz to map the inverse transform to. The DPSS modes
        will be used to interpolate to freq_out if necessary. Default: freq_in.
    filter_half_width : float, optional
        The half-width of the the delay filter window in seconds.
    eigenval_cutoff : float (optional)
        The threshold for eigenvalues of sinc matrix to include DPSS modes for.
    edge_suppression : float (optional)
        The minimum suppression of a tone at the edge of the filter window
        needed to include a DPSS mode.
    nterms : int (optional)
        The number of DPSS modes to include.
    avg_suppression : float (optional)
        The average suppression within a filter window needed to include a
        DPSS mode.

    Returns
    -------
    A : np.ndarray
        The design matrix specifying the inverse transform. That is,
        A @ dpss_coeffs will return the data in frequency domain (interpolated
        to freq_out).

    Raises
    ------
    ValueError :
        If not one and only one of eigenval_cutoff, edge_suppression, nterms,
        and avg_suppression is specified and the rest are None.

    """
    x = np.ravel(freq_out) * 1e6
    kwargs = {
        "eigenval_cutoff": eigenval_cutoff,
        "edge_suppression": edge_suppression,
        "nterms": nterms,
        "avg_suppression": avg_suppression,
    }
    # remove Nones and put the specified value in a list
    kwarg = {key: [val] for key, val in kwargs.items() if val is not None}
    if not len(kwarg) == 1:
        raise ValueError(f"Specify one and only of {list(kwargs.keys())}.")

    A = dpss_operator(
        x,
        filter_centers=[0],
        filter_half_widths=[filter_half_width],
        cache=None,
        **kwarg,
    )[0]
    return A.real


def freq2dpss(data, freq_in, freq_out, design_matrix):
    freq_in = np.ravel(freq_in)
    freq_out = np.ravel(freq_out)
    A = design_matrix[np.isin(freq_out, freq_in)]
    W = np.eye(freq_in.size)
    coeffs = fit_solution_matrix(W, A) @ np.array(data)
    return coeffs


def dpss2freq(dpss_coeffs, design_matrix):
    return design_matrix @ dpss_coeffs



from astropy.coordinates import AltAz, Galactic, EarthLocation, ICRS
from astropy import units
import healpy as hp
from lunarsky import LunarTopo, MCMF, MoonLocation, Time
import numpy as np

from croissant.healpix import healpix2lonlat
from croissant import rotations


def test_topo2radec():
    phi = np.linspace(0, 2 * np.pi, num=360, endpoint=False)
    theta = np.linspace(0, np.pi, num=181)
    loc = EarthLocation(
        lat=40.5 * units.deg, lon=130.2 * units.deg, height=10.3 * units.m
    )
    time = "2022-06-09 16:12:00"
    ra, dec = rotations.topo2radec(theta, phi, time, loc, grid=True)
    assert -90 <= dec.all() < 90
    assert 0 <= ra.all() < 360
    az = phi.reshape(1, -1) * units.rad
    alt = (np.pi / 2 - theta.reshape(-1, 1)) * units.rad
    aa = AltAz(alt=alt, az=az, location=loc, obstime=time)
    icrs = aa.transform_to(ICRS())
    assert np.allclose(ra, icrs.ra.deg.ravel())
    assert np.allclose(dec, icrs.dec.deg.ravel())

    # change phi/theta from coord axes to datapoints
    phi, theta = [ang.ravel() for ang in np.meshgrid(phi, theta)]
    ra, dec = rotations.topo2radec(theta, phi, time, loc, grid=False)
    assert -90 <= dec.all() < 90
    assert 0 <= ra.all() < 360
    assert np.allclose(ra, icrs.ra.deg.ravel())
    assert np.allclose(dec, icrs.dec.deg.ravel())

    # pass time instance to topo2radec
    time = Time(time)
    ra, dec = rotations.topo2radec(theta, phi, time, loc, grid=False)
    assert -90 <= dec.all() < 90
    assert 0 <= ra.all() < 360
    assert np.allclose(ra, icrs.ra.deg.ravel())
    assert np.allclose(dec, icrs.dec.deg.ravel())

    # invert radec2topo
    phi = np.linspace(0, 2 * np.pi, num=360, endpoint=False)
    theta = np.linspace(0, np.pi, num=181)[1:-1]  # remove poles
    phi, theta = [ang.ravel() for ang in np.meshgrid(phi, theta)]
    ra, dec = rotations.topo2radec(theta, phi, time, loc, grid=False)
    theta_, phi_ = rotations.radec2topo(ra, dec, time, loc)
    # handle precision issue
    phi_ = np.where(phi_ >= 2 * np.pi - 1e-10, phi_ - 2 * np.pi, phi_)
    assert np.allclose(theta, theta_)
    assert np.allclose(phi, phi_)


def test_radec2topo():
    ra = np.linspace(0, 360, num=360, endpoint=False)
    dec = np.linspace(90, -90, num=181)
    ra, dec = [ang.ravel() for ang in np.meshgrid(ra, dec)]
    loc = EarthLocation(
        lat=40.5 * units.deg, lon=130.2 * units.deg, height=10.3 * units.m
    )
    time = "2022-06-09 16:12:00"
    theta, phi = rotations.radec2topo(ra, dec, time, loc)
    assert 0 <= theta.all() < np.pi / 2
    assert 0 <= phi.all() < 2 * np.pi
    icrs = ICRS(ra=ra * units.deg, dec=dec * units.deg)
    aa = icrs.transform_to(AltAz(location=loc, obstime=time))
    assert np.allclose(theta, np.pi / 2 - aa.alt.rad)
    assert np.allclose(phi, aa.az.rad)

    # use Time instance
    time = Time(time)
    theta, phi = rotations.radec2topo(ra, dec, time, loc)
    assert 0 <= theta.all() < np.pi / 2
    assert 0 <= phi.all() < 2 * np.pi
    assert np.allclose(theta, np.pi / 2 - aa.alt.rad)
    assert np.allclose(phi, aa.az.rad)

    # invert topo2radec
    ra = np.linspace(0, 360, num=360, endpoint=False)
    dec = np.linspace(90, -90, num=181)
    dec = dec[1:-1]  # remove poles since it will mess with ra
    ra, dec = [ang.ravel() for ang in np.meshgrid(ra, dec)]
    theta, phi = rotations.radec2topo(ra, dec, time, loc)
    ra_, dec_ = rotations.topo2radec(theta, phi, time, loc, grid=False)
    # handle precision issue
    ra_ = np.where(ra_ >= 360 - 1e-10, ra_ - 360, ra_)
    assert np.allclose(dec, dec_)
    assert np.allclose(ra, ra_)


def test_topo2mcmf():
    phi = np.linspace(0, 2 * np.pi, num=360, endpoint=False)
    theta = np.linspace(0, np.pi, num=181)
    loc = MoonLocation(
        lat=40.5 * units.deg, lon=130.2 * units.deg, height=10.3 * units.m
    )
    time = "2022-06-09 16:12:00"
    lon, lat = rotations.topo2mcmf(theta, phi, time, loc, grid=True)
    assert -90 <= lat.all() < 90
    assert 0 <= lon.all() < 360
    az = phi.reshape(1, -1) * units.rad
    alt = (np.pi / 2 - theta.reshape(-1, 1)) * units.rad
    aa = LunarTopo(alt=alt, az=az, location=loc, obstime=time)
    mcmf = aa.transform_to(MCMF())
    assert np.allclose(lon, mcmf.spherical.lon.deg.ravel())
    assert np.allclose(lat, mcmf.spherical.lat.deg.ravel())

    # change phi/theta from coord axes to datapoints
    phi, theta = [ang.ravel() for ang in np.meshgrid(phi, theta)]
    lon, lat = rotations.topo2mcmf(theta, phi, time, loc, grid=False)
    assert -90 <= lat.all() < 90
    assert 0 <= lon.all() < 360
    assert np.allclose(lon, mcmf.spherical.lon.deg.ravel())
    assert np.allclose(lat, mcmf.spherical.lat.deg.ravel())

    # pass time instance to topo2mcmf
    time = Time(time)
    lon, lat = rotations.topo2mcmf(theta, phi, time, loc, grid=False)
    assert -90 <= lat.all() < 90
    assert 0 <= lon.all() < 360
    assert np.allclose(lon, mcmf.spherical.lon.deg.ravel())
    assert np.allclose(lat, mcmf.spherical.lat.deg.ravel())

    # invert mcmf2topo
    phi = np.linspace(0, 2 * np.pi, num=360, endpoint=False)
    theta = np.linspace(0, np.pi, num=181)[1:-1]  # remove poles
    phi, theta = [ang.ravel() for ang in np.meshgrid(phi, theta)]
    lon, lat = rotations.topo2mcmf(theta, phi, time, loc, grid=False)
    theta_, phi_ = rotations.mcmf2topo(lon, lat, time, loc)
    # handle precision issue
    phi_ = np.where(phi_ >= 2 * np.pi - 1e-10, phi_ - 2 * np.pi, phi_)
    assert np.allclose(theta, theta_)
    assert np.allclose(phi, phi_)


def test_mcmf2topo():
    lon = np.linspace(0, 360, num=360, endpoint=False)
    lat = np.linspace(90, -90, num=181)
    lon, lat = [ang.ravel() for ang in np.meshgrid(lon, lat)]
    loc = MoonLocation(
        lat=40.5 * units.deg, lon=130.2 * units.deg, height=10.3 * units.m
    )
    time = "2022-06-09 16:12:00"
    theta, phi = rotations.mcmf2topo(lon, lat, time, loc)
    assert 0 <= theta.all() < np.pi / 2
    assert 0 <= phi.all() < 2 * np.pi
    mcmf = MCMF(
        lon=lon * units.deg,
        lat=lat * units.deg,
        representation_type="spherical",
    )
    aa = mcmf.transform_to(LunarTopo(location=loc, obstime=time))
    assert np.allclose(theta, np.pi / 2 - aa.alt.rad)
    assert np.allclose(phi, aa.az.rad)

    # use Time instance
    time = Time(time)
    theta, phi = rotations.mcmf2topo(lon, lat, time, loc)
    assert 0 <= theta.all() < np.pi / 2
    assert 0 <= phi.all() < 2 * np.pi
    assert np.allclose(theta, np.pi / 2 - aa.alt.rad)
    assert np.allclose(phi, aa.az.rad)

    # invert topo2radec
    lon = np.linspace(0, 360, num=360, endpoint=False)
    lat = np.linspace(90, -90, num=181)
    lat = lat[1:-1]  # remove poles since it will mess with ra
    lon, lat = [ang.ravel() for ang in np.meshgrid(lon, lat)]
    theta, phi = rotations.mcmf2topo(lon, lat, time, loc)
    lon_, lat_ = rotations.topo2mcmf(theta, phi, time, loc, grid=False)
    # handle precision issue
    lon_ = np.where(lon_ >= 360 - 1e-10, lon_ - 360, lon_)
    assert np.allclose(lat, lat_)
    assert np.allclose(lon, lon_)


def test_rot_coords():
    # check that we get expected output in all cases

    # topo -> radec
    theta = np.linspace(0, np.pi, 181)
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    phi, theta = [c.ravel() for c in np.meshgrid(phi, theta)]
    time = Time("2022-06-09 16:12:00")
    loc = EarthLocation(
        lat=40 * units.deg, lon=137 * units.deg, height=0 * units.m
    )
    # expected output (deg)
    exp_ra, exp_dec = rotations.topo2radec(theta, phi, time, loc, grid=False)
    el = 90 - np.rad2deg(theta)
    az = np.rad2deg(phi)
    ra, dec = rotations.rot_coords(
        az, el, "topocentric", "equatorial", time=time, loc=loc, lonlat=True
    )
    assert np.allclose(ra, exp_ra)
    assert np.allclose(dec, exp_dec)

    # radec -> topo
    ra = np.linspace(0, 360, 360, endpoint=False)
    dec = np.linspace(90, -90, 181)
    ra, dec = [c.ravel() for c in np.meshgrid(ra, dec)]
    # expected output (rad)
    exp_theta, exp_phi = rotations.radec2topo(ra, dec, time, loc)
    colat = np.pi / 2 - np.deg2rad(dec)
    lon_rad = np.deg2rad(ra)
    theta, phi = rotations.rot_coords(
        colat, lon_rad, "equatorial", "topocentric", time=time, loc=loc
    )
    assert np.allclose(theta, exp_theta)
    assert np.allclose(phi, exp_phi)

    # radec to topo (not on grid)
    ra, dec = healpix2lonlat(128)
    theta, phi = rotations.radec2topo(ra, dec, time, loc)  # expected output
    za = np.pi / 2 - np.deg2rad(dec)
    az = np.deg2rad(ra)
    th, ph = rotations.rot_coords(
        za, az, "equatorial", "topocentric", time=time, loc=loc, lonlat=False
    )
    assert np.allclose(theta, th)
    assert np.allclose(phi, ph)

    # topo -> mcmf
    theta = np.linspace(0, np.pi, 181)
    phi = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    phi, theta = [c.ravel() for c in np.meshgrid(phi, theta)]
    loc = MoonLocation(
        lat=40 * units.deg, lon=137 * units.deg, height=0 * units.m
    )
    # expected output (deg)
    exp_lon, exp_lat = rotations.topo2mcmf(theta, phi, time, loc, grid=False)
    el = 90 - np.rad2deg(theta)
    az = np.rad2deg(phi)
    lon, lat = rotations.rot_coords(
        az, el, "topocentric", "mcmf", time=time, loc=loc, lonlat=True
    )
    assert np.allclose(lon, exp_lon)
    assert np.allclose(lat, exp_lat)

    # mcmf -> topo
    lon = np.linspace(0, 360, 360, endpoint=False)
    lat = np.linspace(90, -90, 181)
    lon, lat = [c.ravel() for c in np.meshgrid(lon, lat)]
    # expected output (rad)
    exp_theta, exp_phi = rotations.mcmf2topo(lon, lat, time, loc)
    colat = np.pi / 2 - np.deg2rad(lat)
    lon_rad = np.deg2rad(lon)
    theta, phi = rotations.rot_coords(
        colat, lon_rad, "mcmf", "topocentric", time=time, loc=loc
    )
    assert np.allclose(theta, exp_theta)
    assert np.allclose(phi, exp_phi)


def test_get_euler():
    # check that we agree with healpy for galactic -> equatorial
    euler = rotations.get_euler(from_coords="galactic", to_coords="equatorial")
    euler_rot_mat = hp.rotator.get_rotation_matrix(
        euler, deg=False, eulertype="ZYX"
    )[0]
    rot = hp.Rotator(coord=["G", "C"])
    assert np.allclose(euler_rot_mat, rot.mat)

    # equatorial -> galactic
    euler = rotations.get_euler(from_coords="equatorial", to_coords="galactic")
    euler_rot_mat = hp.rotator.get_rotation_matrix(
        euler, deg=False, eulertype="ZYX"
    )[0]
    rot = hp.Rotator(coord=["C", "G"])
    assert np.allclose(euler_rot_mat, rot.mat)


def test_hp_rotate():
    # test that this method agrees with astropy
    rot = rotations.hp_rotate("galactic", "equatorial")
    l, b = 30, 60
    ra, dec = rot(l, b, lonlat=True)
    icrs = Galactic(l=l * units.deg, b=b * units.deg).transform_to(ICRS())
    assert np.isclose(ra % 360, icrs.ra.deg)
    assert np.isclose(dec, icrs.dec.deg)

    # galactic -> mcmf
    time = Time("2022-06-16 17:00:00")
    rot = rotations.hp_rotate("galactic", "mcmf", time=time)
    l, b = 0, 0
    lon, lat = rot(l, b, lonlat=True)
    mcmf = Galactic(l=l * units.deg, b=b * units.deg).transform_to(
        MCMF(obstime=time)
    )
    assert np.isclose(lon % 360, mcmf.spherical.lon.deg)
    assert np.isclose(lat, mcmf.spherical.lat.deg)


def test_rotate_map():
    # pixel weights are not computed for nside = 8
    nside = 8
    npix = hp.nside2npix(nside)
    m = np.arange(npix)
    rm = rotations.rotate_map(
        m, from_coords="galactic", to_coords="equatorial"
    )
    hp_rot = hp.Rotator(coord=["G", "C"])
    hprm = hp_rot.rotate_map_alms(m, use_pixel_weights=False)
    assert np.allclose(rm, hprm)

    # compare for nside 64 which uses pixel weights
    nside = 64
    npix = hp.nside2npix(nside)
    m = np.arange(npix)
    rm = rotations.rotate_map(
        m, from_coords="equatorial", to_coords="galactic"
    )
    hp_rot = hp.Rotator(coord=["C", "G"])
    hprm = hp_rot.rotate_map_alms(m, use_pixel_weights=True)
    assert np.allclose(rm, hprm)
    # without pixel weights it should be different
    hprm = hp_rot.rotate_map_alms(m, use_pixel_weights=False)
    assert not np.allclose(rm, hprm)

    # several maps at once
    f = np.linspace(1, 50, 50).reshape(-1, 1)
    maps = m.reshape(1, -1) * f**3.2
    rm = rotations.rotate_map(
        maps, from_coords="galactic", to_coords="equatorial"
    )
    hp_rot = hp.Rotator(coord=["G", "C"])
    hprm = np.empty((f.size, npix))
    for i, m in enumerate(maps):
        hprm[i] = hp_rot.rotate_map_alms(m, use_pixel_weights=True)
    assert np.allclose(rm, hprm)


def test_rotate_alm():
    lmax = 10
    size = hp.Alm.getsize(lmax)
    alm = np.arange(size) + 5j * np.arange(size) ** 2
    rot_alm = rotations.rotate_alm(
        alm, from_coords="galactic", to_coords="equatorial"
    )
    hp_rot = hp.Rotator(coord=["G", "C"])
    hp_rot_alm = hp_rot.rotate_alm(alm)
    assert np.allclose(rot_alm, hp_rot_alm)

    # multiple sets of alms at once
    alm.shape = (1, size)
    f = np.linspace(1, 50, 50).reshape(-1, 1)
    alms = np.repeat(alm, f.size, axis=0) * f**2
    assert alms.shape == (50, size)
    rot_alms = rotations.rotate_alm(
        alms, from_coords="galactic", to_coords="equatorial"
    )
    hp_rot_alms = np.empty_like(alms)
    for i, a in enumerate(alms):
        hp_rot_alms[i] = hp_rot.rotate_alm(a)
    assert np.allclose(rot_alms, hp_rot_alms)


def test_rot_alm_z():
    lmax = 10
    phi = np.pi / 2
    phase = rotations.rot_alm_z(phi, lmax)
    for ell in range(lmax + 1):
        for emm in range(ell + 1):
            ix = hp.Alm.getidx(lmax, ell, emm)
            assert np.isclose(phase[ix], np.exp(1j * emm * phi))

    # rotate a set of angles
    phi = np.linspace(0, 2 * np.pi, num=361)  # 1 deg spacing
    phase = rotations.rot_alm_z(phi, lmax)
    for ell in range(lmax + 1):
        for emm in range(ell + 1):
            ix = hp.Alm.getidx(lmax, ell, emm)
            assert np.allclose(phase[:, ix], np.exp(1j * emm * phi))

    # check that phi = 0 and phi = 2pi give the same answer
    assert np.allclose(phase[0], phase[-1])
